#ifndef _VELIB_QT_VE_QITEM_HPP_
#define _VELIB_QT_VE_QITEM_HPP_

#include <QDebug>
#include <QList>
#include <QObject>
#include <QString>
#include <QVariant>

class VeQItem;
class VeQItemProducer;

/*
 * Just a helper to invoke a slots as callback functions.
 * Should actually be hidden, but needs to be in some header file
 * for moc to see it, so it lives here for now.
 */
class VeQItemForeach : public QObject {
	Q_OBJECT

public:
	// note: member is SLOT(xyz) in obj to be invoke from handleItem.
	VeQItemForeach(QObject *obj, const char *member, void *ctx) :
		mCtx(ctx)
	{
		if (member)
			obj->connect(this, SIGNAL(doHandleItem(VeQItem*,void*)), member, Qt::DirectConnection);
	}

	virtual void handleItem(VeQItem *item)
	{
		emit doHandleItem(item, mCtx);
	}

signals:
	void doHandleItem(VeQItem *item, void *ctx);

protected:
	void *mCtx;
};

/**
 * Base class for an item. An item can hold any information about a
 * device / system or also settings. It can be associated with a local device,
 * bus or something at the other end of the world with many interfaces in
 * between. We don't care here, it is left to the producer to take care of that.
 * The root of the tree is special in that sense, it is the only item which is not
 * produced (and hence producer of the root is 0), see VeQItems.
 */
class VeQItem : public QObject
{
	Q_OBJECT
	Q_PROPERTY(QVariant value READ getValue NOTIFY valueChanged WRITE setValue)
	Q_PROPERTY(QVariant text READ getText NOTIFY valueChanged)
	Q_PROPERTY(QString uid READ uniqueId CONSTANT)
	Q_PROPERTY(QString id READ id CONSTANT)

public:
	typedef QMap<QString, VeQItem *> Children;

	enum State {
		Idle,
		Offline,
		Requested,
		Storing,
		Synchronized
	};

	explicit VeQItem(VeQItemProducer *producer, QObject *parent = 0);
	~VeQItem();

	/**
	 * @brief request for the item its value.
	 */
	virtual QVariant getValue()
	{
		if (mState == Idle)
			setState(Requested);
		return mValue;
	}

	/**
	 * Invokes the slot member with the current value, and future onChangeEvents.
	 */
	void getValueAndChanges(QObject *obj, const char *member);

	/**
	 * Like getValue, but a human representable version.
	 */
	virtual QString getText()
	{
		return mText;
	}

	/**
	 * Like getValue, in general no assumption can be made about the time
	 * it takes to make a change. This defaut implementation just set the
	 * value directly though, it is expected to be overridden.
	 */
	virtual int setValue(QVariant const &value)
	{
		produceValue(value);
		return 0;
	}

	Children const &itemChildren() const { return mChilds; }
	State getState() { return mState; }
	State getTextState() { return mTextState; }
	bool isLeaf() { return mIsLeaf; }

	virtual void produceValue(QVariant value, State state = Synchronized);
	virtual void produceText(QString text, State state = Synchronized);

	virtual VeQItem *createChild(QString id, bool isLeaf = true);
	VeQItem *createChild(QString id, QVariant var);

	QString id();
	void setId(QString id);
	QString uniqueId();

	// note the Item prefixes are a bit verbose, but prevents name collision with QObjects
	// own child, parent member functions.
	VeQItem *itemChild(int n);
	Q_INVOKABLE VeQItem *itemGet(QString uid);
	Q_INVOKABLE VeQItem *itemGetOrCreate(QString uid, bool isLeaf = true);
	Q_INVOKABLE VeQItem *itemGetOrCreateAndProduce(QString uid, QVariant value);

	VeQItem *itemRoot();
	VeQItem *itemParent();

	/**
	 * Additional / optional properties, like min, max, defaultValue
	 * property("min") etc will return them or an invalid variant when non existing.
	 */
	virtual QVariant itemProperty(const char *name);
	virtual void itemProduceProperty(const char *name, const QVariant &value, State state = Synchronized);

	VeQItemProducer *producer() { return mProducer; }
	void setProducer(VeQItemProducer *producer) { mProducer = producer; }

	int index();

	void foreachChildFirst(QObject *obj, const char *member, void *ctx = 0);
	void foreachChildFirst(VeQItemForeach *each);

	void foreachParentFirst(QObject *obj, const char *member, void *ctx = 0);
	void foreachParentFirst(VeQItemForeach *each);

	void itemAddChild(QString id, VeQItem *item);
	void itemDeleteChild(VeQItem *child);
	void itemDelete();

	/// Returns the relative path from `ancestor` to `this`.
	/// @returns The relative path (for now, starting with a slash). If `ancestor`
	/// is not an ancestor of `this` an empty string will be returned.
	QString getRelId(VeQItem *ancestor);

signals:
	void dynamicPropertyChanged(VeQItem *item, char const *name, QVariant value);
	void valueChanged(VeQItem *item, QVariant var);
	void textChanged(VeQItem *item, QString text);
	void stateChanged(VeQItem *item, State state);
	void textStateChanged(VeQItem *item, State state);

	/*
	 * Note: if you want to connect to signals of the childs in a cross thread
	 * manner, a connection should be made with Qt::BlockingQueuedConnection.
	 * Perhaps such use should simply be prohibited.
	 */
	void childAboutToBeAdded(VeQItem *item);
	void childAdded(VeQItem *item);
	void childAboutToBeRemoved(VeQItem *item);
	void childRemoved(VeQItem *item);

protected:
	using QObject::disconnectNotify; // tell the compiler we want both, this one and QObject's
	virtual void disconnectNotify(const char *signal);
	virtual void watchedChanged();
	// by this time all signals can be assumed to be hooked up..
	virtual void afterAdd();
	void setState(State state);
	void setTextState(State state);
	// last point before the item is announced
	virtual void setParent(QObject *parent);

protected slots:
	void receiverDestroyed(QObject *obj);
	void resetId(VeQItem *item, void *ctx);

private:
	void uniqueId(QString &uid);
	void updateWatched();

protected:
	QString mId;
	Children mChilds;
	QVariant mValue;
	State mState;
	QString mText;
	State mTextState;
	VeQItemProducer *mProducer;
	bool mIsLeaf;
	QString mUid;
	bool mWatched;
	QHash<QString, State> mPropertyState;
};

/* Singleton to get the root item */
class VeQItems
{
public:
	static VeQItem *getRoot()
	{
		static VeQItem theRoot(0);
		return &theRoot;
	}
};

class VeQuickItem : public QObject {
	Q_OBJECT
	Q_PROPERTY(QVariant defaultValue READ getDefault NOTIFY defaultChanged)
	Q_PROPERTY(QVariant min READ getMin NOTIFY minChanged)
	Q_PROPERTY(QVariant max READ getMax NOTIFY maxChanged)
	Q_PROPERTY(QVariant text READ getText NOTIFY textChanged)
	Q_PROPERTY(QString uid READ getUid NOTIFY uidChanged WRITE setUid)
	Q_PROPERTY(QVariant value READ getValue NOTIFY valueChanged WRITE setValueProperty)

public:
	VeQuickItem() : QObject(),
		mItem(0)
	{
		setUid("");
	}

	~VeQuickItem()
	{
		teardown();
	}

	QVariant getDefault() { return mItem->itemProperty("defaultValue"); }
	QVariant getMax() { return mItem->itemProperty("max"); }
	QVariant getMin() { return mItem->itemProperty("min"); }

	QString getText() { return mItem->getText(); }
	QString getUid() { return mItem->uniqueId(); }
	void setUid(QString uid);
	QVariant getValue() { return mItem->getValue(); }
	Q_INVOKABLE int setValue(QVariant const &value) { return mItem->setValue(value); }
	void setValueProperty(QVariant value);

signals:
	void defaultChanged();
	void minChanged();
	void maxChanged();

	void textChanged(VeQItem *item, QString text);
	void uidChanged();
	void valueChanged(VeQItem *item, QVariant var);

private slots:
	void onDynamicPropertyChanged(VeQItem *item, char const *name)
	{
		Q_UNUSED(item);

		if (QLatin1String(name) == "min")
			emit minChanged();
		else if (QLatin1String(name) == "max")
			emit maxChanged();
		else if (QLatin1String(name) == "defaultValue")
			emit defaultChanged();
	}

private:
	void setup()
	{
		mItem->getValueAndChanges(this, SIGNAL(valueChanged(VeQItem*,QVariant)));

		connect(mItem, SIGNAL(textChanged(VeQItem*,QString)), SIGNAL(textChanged(VeQItem*,QString)));
		emit textChanged(mItem, mItem->getText());

		connect(mItem, SIGNAL(dynamicPropertyChanged(VeQItem*,const char*,QVariant)), SLOT(onDynamicPropertyChanged(VeQItem*,char const *)));
		emit minChanged();
		emit maxChanged();
		emit defaultChanged();
	}

	void teardown()
	{
		if (mItem == 0)
			return;

		mItem->disconnect(this);
		if (mItem->uniqueId() == "")
			delete mItem;
		mItem = 0;
	}

	VeQItem *mItem;
};

/*
 * Base class for a provider of items. This is almost the same as
 * a dbus service, but a ItemProvider can contain more then one dbus
 * service. E.g. a remote CCGX connection allowing accesses to the dbus
 * provides all dbus services of the remote side at once.
 */
class VeQItemProducer : public QObject
{
	Q_OBJECT

public:

	/* Note: the root item must exists for the live time of the item producers */
	VeQItemProducer(VeQItem *root, QString id, QObject *parent = 0) :
		QObject(parent),
		mProducerRoot(createItem())
	{
		root->itemAddChild(id, mProducerRoot);
	}

	virtual bool open() { return true; }

	// factory function for items
	virtual VeQItem *createItem() { return new VeQItem(this); }

	/*
	 * Get item where services are added.
	 * by default services register in the root item of its producer ..
	 */
	virtual VeQItem *services() { return mProducerRoot; }

protected:
	VeQItem *mProducerRoot;
};

#endif
