#include <QDebug>
#include <QMetaObject>
#include <QStringList>

#include <velib/qt/ve_qitem.hpp>

VeQItem::VeQItem(VeQItemProducer *producer, QObject *parent) :
	QObject(parent),
	mState(Idle),
	mTextState(Idle),
	mProducer(producer),
	mIsLeaf(false),
	mWatched(false)
{
}

VeQItem::~VeQItem() {
	foreach (VeQItem *child, mChilds)
		itemDeleteChild(child);
}

void VeQItem::setParent(QObject *parent)
{
	QObject::setParent(parent);
	foreachParentFirst(this, SLOT(resetId(VeQItem *, void *)), 0);
}

void VeQItem::getValueAndChanges(QObject *obj, const char *member)
{
	QString method(&member[1]);
	method = method.left(method.indexOf("("));
	connect(this, SIGNAL(valueChanged(VeQItem*,QVariant)), obj, member);
	updateWatched();
	QMetaObject::invokeMethod(obj, method.toLatin1(), Q_ARG(VeQItem*, this), Q_ARG(QVariant, getValue()));

	// there is BUG in qt 4, see https://bugreports.qt.io/browse/QTBUG-4844
	// which causes disconnectNotify not to be called when the receiver is deleted.
	// Hence monitor such deletions and explicitly disconnect the receiver upon destruction.
	connect(obj, SIGNAL(destroyed(QObject*)), SLOT(receiverDestroyed(QObject*)));
}

void VeQItem::updateWatched()
{
	bool watched = receivers(SIGNAL(valueChanged(VeQItem*,QVariant))) != 0;
	if (mWatched == watched)
		return;
	mWatched = watched;
	watchedChanged();
}

void VeQItem::receiverDestroyed(QObject *obj)
{
	disconnect(obj);
}

void VeQItem::disconnectNotify(const char *signal)
{
	Q_UNUSED(signal);
	updateWatched();
}

/**
 * The mWatched member keeps track if there is a receiver interested in value
 * changes at all. In such cases there is no need to refresh the value continuesly
 * since it will be ignored anyway. The watchedChanged can be overridden by a
 * producer.
 *
 * Warning: This function violates the object-oriented principle of modularity.
 * However, it might be useful when you need to perform expensive initialization
 * only if something is connected to a signal.
 */
void VeQItem::watchedChanged()
{
}

VeQItem *VeQItem::itemChild(int n)
{
	if (n >= mChilds.count())
		return 0;

	// check if this is expensive
	return mChilds[mChilds.keys()[n]];
}

void VeQItem::itemAddChild(QString id, VeQItem *item)
{
	item->setId(id);
	item->setParent(this);
	mIsLeaf = false;
	emit childAboutToBeAdded(item);
	mChilds[id] = item;
	emit childAdded(item);
	item->afterAdd();
}

void VeQItem::afterAdd()
{
}

void VeQItem::itemDeleteChild(VeQItem *child)
{
	emit childAboutToBeRemoved(child);
	mChilds.remove(child->mId);
	emit childRemoved(child);
	child->deleteLater();
}

// deletes the item and removes it from its parent
void VeQItem::itemDelete()
{
	itemParent()->itemDeleteChild(this);
}

QString VeQItem::getRelId(VeQItem *ancestor)
{
	Q_ASSERT(ancestor != 0);
	if (this == ancestor)
		return "/";
	VeQItem *parent = itemParent();
	if (parent == 0)
		return QString();
	QString p = parent->getRelId(ancestor);
	if (p.isEmpty())
		return p;
	if (!p.endsWith('/'))
		p += '/';
	p += id();
	return p;
}

VeQItem *VeQItem::createChild(QString id, QVariant var)
{
	VeQItem *item = createChild(id);
	item->produceValue(var);
	return item;
}

VeQItem *VeQItem::createChild(QString id, bool isLeaf)
{
	Q_ASSERT(producer());

	VeQItem *item = producer()->createItem();
	item->mIsLeaf = isLeaf;
	itemAddChild(id, item);
	return item;
}

void VeQItem::produceValue(QVariant variant, State state)
{
	bool stateIsChanged = mState != state;
	bool valueIsChanged = mValue != variant;

	mState = state;
	mValue = variant;

	if (stateIsChanged)
		emit stateChanged(this, state);
	if (valueIsChanged)
		emit valueChanged(this, variant);
}

void VeQItem::produceText(QString text, VeQItem::State state)
{
	bool stateIsChanged = mTextState != state;
	bool textIsChanged = mText != text;

	mTextState = state;
	mText = text;

	if (stateIsChanged)
		emit textStateChanged(this, state);
	if (textIsChanged)
		emit textChanged(this, text);
}

QString VeQItem::id()
{
	return mId;
}

void VeQItem::setId(QString id)
{
	Q_ASSERT(parent() == 0);

	mId = id;
	setObjectName(id);
}

QString VeQItem::uniqueId()
{
	if (!mUid.isNull())
		return mUid;

	QString ret;
	uniqueId(ret);
	return ret;
}

void VeQItem::uniqueId(QString &uid)
{
	VeQItem *theParent = qobject_cast<VeQItem *>(parent());
	if (theParent) {
		theParent->uniqueId(uid);
		uid += (uid == "" ? "" : "/") + mId;
	} else {
		uid = mId;
	}
}

void VeQItem::resetId(VeQItem *item, void *ctx)
{
	Q_UNUSED(ctx);
	QString str;
	item->uniqueId(str);
	item->mUid = str;
}

VeQItem *VeQItem::itemGet(QString uid)
{
	VeQItem *item = this;

	// tolerate ids starting with a slash
	if (uid.startsWith("/"))
		uid = uid.mid(1);

	QStringList parts = uid.split('/');
	foreach (QString part, parts) {
		if (part == "") {
			item = itemRoot();
		} else {
			item = item->itemChildren()[part];
			if (item == 0)
				return 0;
		}
	}
	return item;
}

VeQItem *VeQItem::itemGetOrCreate(QString uid, bool isLeaf)
{
	VeQItem *item = this;

	// tolerate ids starting with a slash
	if (uid.startsWith("/"))
		uid = uid.mid(1);

	QStringList parts = uid.split('/');
	int n = 1;
	foreach (QString part, parts) {
		VeQItem *child = item->itemChildren()[part];
		if (child == 0)
			child = item->createChild(part, isLeaf && n == parts.count());
		item = child;
		n++;
	}
	return item;
}

VeQItem *VeQItem::itemGetOrCreateAndProduce(QString uid, QVariant value)
{
	VeQItem *item = itemGetOrCreate(uid);
	item->produceValue(value);
	return item;
}

VeQItem *VeQItem::itemRoot()
{
	VeQItem *ret = this;
	for (;;) {
		VeQItem *parent = ret->itemParent();
		if (!parent)
			return ret;
		ret = parent;
	}
}

VeQItem *VeQItem::itemParent()
{
	return qobject_cast<VeQItem *>(parent());
}

QVariant VeQItem::itemProperty(const char *name)
{
	if (mPropertyState[name] != Synchronized)
		mPropertyState[name] = Requested;
	return property(name);
}

void VeQItem::itemProduceProperty(const char *name, const QVariant &value, VeQItem::State state)
{
	bool changed = mPropertyState[name] != state;
	mPropertyState[name] = state;
	setProperty(name, value);
	if (changed)
		emit dynamicPropertyChanged(this, name, value);
}

// returns the index in the parents it child ids.
int VeQItem::index()
{
	VeQItem *theParent = qobject_cast<VeQItem *>(parent());
	if (!theParent)
		return 0;

	return theParent->itemChildren().keys().indexOf(mId);
}

void VeQItem::foreachChildFirst(VeQItemForeach *each)
{
	foreach (VeQItem *child, mChilds)
		child->foreachChildFirst(each);
	each->handleItem(this);
}

void VeQItem::foreachChildFirst(QObject *obj, const char *member, void *ctx)
{
	VeQItemForeach each(obj, member, ctx);
	foreachChildFirst(&each);
}

void VeQItem::foreachParentFirst(VeQItemForeach *each)
{
	each->handleItem(this);
	foreach (VeQItem *child, mChilds)
		child->foreachParentFirst(each);
}

void VeQItem::foreachParentFirst(QObject *obj, const char *member, void *ctx)
{
	VeQItemForeach each(obj, member, ctx);
	foreachParentFirst(&each);
}

void VeQItem::setState(VeQItem::State state)
{
	if (mState == state)
		return;
	mState = state;
	emit stateChanged(this, state);
}

void VeQItem::setTextState(VeQItem::State state)
{
	if (mTextState == state)
		return;
	mTextState = state;
	emit textStateChanged(this, state);
}

/* QT Quick Part */

void VeQuickItem::setUid(QString uid)
{
	if (mItem) {
		// check it changed at all
		if (uid == mItem->uniqueId())
			return;

		if (mItem->uniqueId().isEmpty() && mItem->getValue().isValid())
			qDebug() << "Changing an uid of an item with valid value set is weird, ignoring value" <<
						mItem->getValue() << "for" << uid;
	}

	teardown();
	if (uid == "")
		mItem = new VeQItem(0);
	else
		mItem = VeQItems::getRoot()->itemGetOrCreate(uid);
	setup();
	emit uidChanged();
}

/**
 * Handle changes of the value property.
 *
 * In qml it should be possible to set the value property of local items (the once
 * "produced"), but not the once being bound (the once "consumed"). In the latter
 * case the local value would otherwise be updated before it is even send to the remote
 * side (which might refuse the change, or not even be available at the moment etc). In
 * such a case the value property should be updated after the value was actually changed
 * and confirmed to us, which is exactly what setValue does and that should be used
 * instead in such cases (and will act exactly the same on local items).
 *
 * For example:
 *
 * VeQuickItem {
 *   value: "This should work fine!"
 * }
 *
 * VeQuickItem {
 *   value: "This should be refused"
 *   bind: "Some/Remote/Value"
 * }
 *
 * The latter should be changed to:
 *
 *  VeQuickItem {
 *   bind: "Some/Remote/Value"
 *
 *   Component.onCompleted: setValue("This should work as well")
 *   onValueChanged: console.log("the remote value changed to " + value)
 * }
 */
void VeQuickItem::setValueProperty(QVariant value)
{
	if (mItem->uniqueId() != "") {
		qDebug() << "ignoring request to set value on bound qml Item, please use setValue instead";
		return;
	}

	mItem->produceValue(value);
	mItem->produceText(value.toString());
}
