#include <QDBusArgument>
#include <QDBusMessage>
#include <QDBusVariant>
#include <QDebug>
#include <velib/qt/ve_qitem.hpp>
#include "ve_qitem_dbus_virtual_object.hpp"

Q_DECLARE_METATYPE(QList<int>)

VeQItemDbusVirtualObject::VeQItemDbusVirtualObject(const QDBusConnection &connection, VeQItem *root,
												   QObject *parent):
	QDBusVirtualObject(parent),
	mConnection(connection),
	mRoot(root)
{
	Q_ASSERT(root != 0);
	connectItem(root);
}

VeQItem *VeQItemDbusVirtualObject::root()
{
	return mRoot;
}

QString VeQItemDbusVirtualObject::serviceName() const
{
	return mConnection.name();
}

bool VeQItemDbusVirtualObject::registerService()
{
	if (!mConnection.registerVirtualObject("/", this, QDBusConnection::SubPath)) {
		qDebug() << "[VeQItemDbusVirtualObject] Could not register virtual object";
		return false;
	}
	if (!mConnection.registerService(serviceName())) {
		qDebug() << "[VeQItemDbusVirtualObject] Could not register service" << serviceName();
		return false;
	}
	return true;
}

bool VeQItemDbusVirtualObject::unregisterService()
{
	if (!mConnection.unregisterService(serviceName())) {
		qDebug() << "[VeQItemDbusVirtualObject] Could not unregister service" << serviceName();
		return false;
	}
	return true;
}

QString VeQItemDbusVirtualObject::introspect(const QString &path) const
{
	static const QString Interface =
		"  <interface name=\"com.victronenergy.BusItem\">\n"
		"    <method name=\"GetDescription\">\n"
		"      <arg direction=\"in\" type=\"s\" name=\"language\"/>\n"
		"      <arg direction=\"in\" type=\"i\" name=\"length\"/>\n"
		"      <arg direction=\"out\" type=\"s\" name=\"descr\"/>\n"
		"    </method>\n"
		"    <method name=\"GetValue\">\n"
		"      <arg direction=\"out\" type=\"v\" name=\"value\"/>\n"
		"    </method>\n"
		"    <method name=\"GetText\">\n"
		"      <arg direction=\"out\" type=\"s\" name=\"value\"/>\n"
		"    </method>\n"
		"    <method name=\"SetValue\">\n"
		"      <arg direction=\"in\" type=\"v\" name=\"value\"/>\n"
		"      <arg direction=\"out\" type=\"i\" name=\"retval\"/>\n"
		"    </method>\n"
		"    <method name=\"GetMin\">\n"
		"      <arg direction=\"out\" type=\"v\" name=\"value\"/>\n"
		"    </method>\n"
		"    <method name=\"GetMax\">\n"
		"      <arg direction=\"out\" type=\"v\" name=\"value\"/>\n"
		"    </method>\n"
		"    <property access=\"readwrite\" type=\"v\" name=\"Value\"/>\n"
		"    <property access=\"read\" type=\"s\" name=\"Text\"/>\n"
		"    <signal name=\"PropertiesChanged\">\n"
		"      <arg direction=\"out\" type=\"a{sv}\" name=\"changes\"/>\n"
		"      <annotation value=\"QVariantMap\" name=\"com.trolltech.QtDBus.QtTypeName.In0\"/>\n"
		"    </signal>\n"
		"  </interface>\n";
	QString result = Interface;
	VeQItem *item = findItem(path);
	if (item == 0)
		return QString();
	for (int i=0;;++i) {
		VeQItem *child = item->itemChild(i);
		if (child == 0)
			break;
		result.append(QString("  <node name=\"%1\"/>\n").arg(child->id()));
	}
	return result;
}

bool VeQItemDbusVirtualObject::handleMessage(const QDBusMessage &message,
											 const QDBusConnection &connection)
{
	switch (message.type()) {
	case QDBusMessage::MethodCallMessage:
	{
		QString member = message.member();
		VeQItem *item = findItem(message.path());
		if (item == 0) {
			return false;
		} else {
			if (member == "GetValue") {
				return handleGetValue(message, connection, item);
			} else if (member == "GetText") {
				return handleGetText(message, connection, item);
			} else if (member == "SetValue") {
				return handleSetValue(message, connection, item);
			}
		}
		break;
	}
	default:
		qDebug() << "[VeQItemDbusVirtualObject] unhandled message type:" << message.type();
		break;
	}
	return false;
}

bool VeQItemDbusVirtualObject::handleGetValue(const QDBusMessage &message,
											  const QDBusConnection &connection, VeQItem *item)
{
	if (item->isLeaf()) {
		QVariant v = item->getValue();
		denormalizeVariant(v);
		QDBusMessage reply = message.createReply(QVariant::fromValue(QDBusVariant(v)));
		return connection.send(reply);
	} else {
		QVariantMap map;
		buildTree(item, item, map);
		QDBusMessage reply = message.createReply(QVariant::fromValue(QDBusVariant(map)));
		return connection.send(reply);
	}
}

bool VeQItemDbusVirtualObject::handleGetText(const QDBusMessage &message,
											 const QDBusConnection &connection, VeQItem *item)
{
	if (item->isLeaf()) {
		QDBusMessage reply = message.createReply(QVariant::fromValue(item->getText()));
		return connection.send(reply);
	} else {
		QVariantMap map;
		buildTree(item, item, map, true);
		QDBusMessage reply = message.createReply(QVariant::fromValue(QDBusVariant(map)));
		return connection.send(reply);
	}
}

bool VeQItemDbusVirtualObject::handleSetValue(const QDBusMessage &message,
											  const QDBusConnection &connection, VeQItem *producer)
{
	QList<QVariant> args = message.arguments();
	if (args.size() != 1) {
		QDBusMessage reply = message.createErrorReply(QDBusError::InvalidArgs,
													  "Expected 1 argument");
		return connection.send(reply);
	}
	QVariant v = args.first();
	normalizeVariant(v);
	int r = producer->setValue(v);
	QDBusMessage reply = message.createReply(r);
	return connection.send(reply);
}

void VeQItemDbusVirtualObject::buildTree(VeQItem *root, VeQItem *item, QVariantMap &map,
										 bool useText)
{
	for (int i=0;;++i) {
		VeQItem *child = item->itemChild(i);
		if (child == 0)
			break;
		if (child->isLeaf()) {
			QString path = child->getRelId(root);
			if (path.startsWith('/'))
				path.remove(0, 1);
			if (useText) {
				map[path] = child->getText();
			} else {
				QVariant v = child->getValue();
				denormalizeVariant(v);
				map[path] = v;
			}
		} else {
			buildTree(root, child, map, useText);
		}
	}
}

void VeQItemDbusVirtualObject::addPending(VeQItem *item)
{
	if (!mPendingChanges.contains(item)) {
		if (mPendingChanges.isEmpty())
			QMetaObject::invokeMethod(this, "processPending", Qt::QueuedConnection);
		mPendingChanges.append(item);
	}
}

void VeQItemDbusVirtualObject::processPending()
{
	QList<VeQItem *> pending = mPendingChanges;
	mPendingChanges.clear();
	foreach (VeQItem *p, pending)
		sendPropertiesChanged(p);
}

void VeQItemDbusVirtualObject::sendPropertiesChanged(VeQItem *item)
{
	QString dbusPath = item->getRelId(mRoot);
	QDBusMessage message = QDBusMessage::createSignal(dbusPath, "com.victronenergy.BusItem",
													  "PropertiesChanged");
	QVariantMap map;
	QVariant v = item->getValue();
	denormalizeVariant(v);
	map.insert("Value", v);
	map.insert("Text", item->getText());
	message << map;
	if (!mConnection.send(message))
		qDebug() << "[VeQItemDbusVirtualObject] Could not send PropertiesChanged signal on path"
				 << dbusPath;
}

void VeQItemDbusVirtualObject::connectItem(VeQItem *item)
{
	if (item->isLeaf()) {
		connect(item, SIGNAL(valueChanged(VeQItem *, QVariant)),
				this, SLOT(onValueChanged(VeQItem *)));
		connect(item, SIGNAL(textChanged(VeQItem *, QString)),
				this, SLOT(onTextChanged(VeQItem *)));
	} else {
		connect(item, SIGNAL(childAdded(VeQItem *)), this, SLOT(onChildAdded(VeQItem *)));
		connect(item, SIGNAL(childRemoved(VeQItem *)), this, SLOT(onChildRemoved(VeQItem *)));
		for (int i=0;;++i) {
			VeQItem *child = item->itemChild(i);
			if (child == 0)
				break;
			connectItem(child);
		}
	}
}

void VeQItemDbusVirtualObject::disconnectItem(VeQItem *item)
{
	if (item->isLeaf()) {
		disconnect(item, SIGNAL(valueChanged(VeQItem *, QVariant)),
				   this, SLOT(onValueChanged(VeQItem *)));
		disconnect(item, SIGNAL(textChanged(VeQItem *, QString)),
				   this, SLOT(onTextChanged(VeQItem *)));
	} else {
		disconnect(item, SIGNAL(childAdded(VeQItem *)), this, SLOT(onChildAdded(VeQItem *)));
		disconnect(item, SIGNAL(childRemoved(VeQItem *)), this, SLOT(onChildRemoved(VeQItem *)));
		for (int i=0;;++i) {
			VeQItem *child = item->itemChild(i);
			if (child == 0)
				break;
			disconnectItem(child);
		}
	}
}

void VeQItemDbusVirtualObject::onChildAdded(VeQItem *item)
{
	connectItem(item);
}

void VeQItemDbusVirtualObject::onChildRemoved(VeQItem *item)
{
	disconnectItem(item);
}

void VeQItemDbusVirtualObject::onValueChanged(VeQItem *item)
{
	addPending(item);
}

void VeQItemDbusVirtualObject::onTextChanged(VeQItem *item)
{
	addPending(item);
}

VeQItem *VeQItemDbusVirtualObject::findItem(const QString &path) const
{
	if (path == "/")
		return mRoot;
	return mRoot->itemGet(path);
}

void VeQItemDbusVirtualObject::normalizeVariant(QVariant &v)
{
	if (v.userType() == QVariant::Map) {
		QVariantMap m = v.toMap();
		normalizeMap(m);
	} else if (v.userType() == qMetaTypeId<QDBusVariant>()) {
		v = qvariant_cast<QDBusVariant>(v).variant();
		normalizeVariant(v);
	} else if (v.userType() == qMetaTypeId<QDBusArgument>()) {
		QDBusArgument arg = qvariant_cast<QDBusArgument>(v);
		if (arg.currentSignature() == "a{sv}") {
			QVariantMap m = qdbus_cast<QVariantMap>(arg);
			normalizeMap(m);
			v = m;
		} else  if (arg.currentSignature() == "av") {
			QVariantList vl = qdbus_cast<QVariantList>(arg);
			normalizeList(vl);
			v = vl;
		} else  if (arg.currentSignature() == "ai") {
			// An empty list of integers is used by victron to encode an invalid
			// value, because D-Bus itself does not define an invalid (or null)
			// value.
			QList<int> vl = qdbus_cast<QList<int> >(arg);
			if (vl.size() == 0)
				v = QVariant();
		} else {
			qDebug() <<	"Cannot handle signature:" << arg.currentSignature();
		}
	}
}

void VeQItemDbusVirtualObject::normalizeMap(QVariantMap &map)
{
	for (QVariantMap::iterator it = map.begin(); it != map.end(); ++it)
		normalizeVariant(it.value());
}

void VeQItemDbusVirtualObject::normalizeList(QVariantList &list)
{
	for (QVariantList::iterator it = list.begin(); it != list.end(); ++it)
		normalizeVariant(*it);
}

void VeQItemDbusVirtualObject::denormalizeVariant(QVariant &v)
{
	if (!v.isValid())
		v = QVariant::fromValue(QList<int>());
}
