#include <QDebug>

#include <velib/qt/ve_qitem.hpp>
#include <velib/qt/ve_qitem_table_model.hpp>

VeQItemTableModel::VeQItemTableModel(int flags, QObject *parent) :
	QAbstractItemModel(parent),
	mFlags(flags)
{
	mColumns << "id";
	mColumns << "value" << "state";
	if (mFlags & VeQItemModelWithText)
		mColumns << "text" << "textState";
	if (mFlags & VeQItemModelWithSettingInfo)
		mColumns << "min" << "max" << "defaultValue";

	// qt 5 has roleNames as virtual and setRoleNames is removed
#if QT_VERSION < QT_VERSION_CHECK(5, 0, 0)
	setRoleNames(roleNames());
#endif
}

void VeQItemTableModel::addItem(VeQItem *item)
{
	if (!item)
		return;

	if (mMap.contains(item->uniqueId())) {
		qWarning() << item << " " << item->uniqueId() << " was already added";
		return;
	}

	if (mFlags & VeQItemModelAddAllChildren) {
		// Add this, all children and their children
		if ((mFlags & VeQItemModelDontAddItem) == 0)
			setupValueChanges(item, VeQItemModelAddAllChildren);
		foreach (VeQItem *child, item->itemChildren())
			setupValueChanges(child, VeQItemModelAddAllChildren);

	} else if (mFlags & VeQItemModelAddChildren) {
		// Only add this and direct siblings
		if ((mFlags & VeQItemModelDontAddItem) == 0)
			setupValueChanges(item, VeQItemModelAddChildren);
		foreach (VeQItem *child, item->itemChildren())
			setupValueChanges(child);

	} else {
		// only add this item
		setupValueChanges(item);
	}
}

int VeQItemTableModel::columnCount(const QModelIndex &parent) const
{
	Q_UNUSED(parent);

	return mColumns.count();
}

void VeQItemTableModel::cellChanged(VeQItem *item, QString column)
{
	int n = mMap.values().indexOf(item);
	int i = mColumns.indexOf(column);
	QModelIndex from = createIndex(n, i, item);
	QModelIndex till = createIndex(n, i, item);
	emit dataChanged(from, till);
}

void VeQItemTableModel::onValueChanged(VeQItem *item)
{
	cellChanged(item, "value");
}

void VeQItemTableModel::onStateChanged(VeQItem *item)
{
	cellChanged(item, "state");
}

void VeQItemTableModel::onTextChanged(VeQItem *item)
{
	cellChanged(item, "text");
}

void VeQItemTableModel::onTextStateChanged(VeQItem *item)
{
	cellChanged(item, "textState");
}

void VeQItemTableModel::onDynamicPropertyChanged(VeQItem *item, const char *name)
{
	cellChanged(item, name);
}

void VeQItemTableModel::setupValueChanges(VeQItem *item, VeQItemModelOptions options)
{
	if (options == VeQItemModelAddAllChildren)
		this->connect(item, SIGNAL(childAdded(VeQItem *)), SLOT(onRecursiveChildAdded(VeQItem *)));
	else if (options == VeQItemModelAddChildren)
		this->connect(item, SIGNAL(childAdded(VeQItem *)), SLOT(onChildAdded(VeQItem *)));

	this->connect(item, SIGNAL(childAboutToBeRemoved(VeQItem*)), SLOT(onItemAboutToBeRemoved(VeQItem*)));

	// flatten the tree for the table view, only interested in the leafs..
	if (!item->isLeaf() && (mFlags & VeQItemModelAddNoneLeaves) == 0)
		return;

	this->connect(item, SIGNAL(valueChanged(VeQItem*,QVariant)), SLOT(onValueChanged(VeQItem*)));
	this->connect(item, SIGNAL(stateChanged(VeQItem*,State)), SLOT(onStateChanged(VeQItem*)));
	this->connect(item, SIGNAL(textChanged(VeQItem*,QString)), SLOT(onTextChanged(VeQItem*)));
	this->connect(item, SIGNAL(textStateChanged(VeQItem*,State)), SLOT(onTextStateChanged(VeQItem*)));
	this->connect(item, SIGNAL(dynamicPropertyChanged(VeQItem*,const char*,QVariant)), SLOT(onDynamicPropertyChanged(VeQItem*,const char*)));

	QString key(item->uniqueId());
	mMap.insert(key, item);
}

void VeQItemTableModel::addExistingChildren(VeQItem *item, void *ctx)
{
	Q_UNUSED(ctx);
	setupValueChanges(item, VeQItemModelAddAllChildren);
}


void VeQItemTableModel::onRecursiveChildAdded(VeQItem *item)
{
	item->foreachParentFirst(this, SLOT(addExistingChildren(VeQItem *, void *)), 0);
	emit layoutChanged();
}

void VeQItemTableModel::onChildAdded(VeQItem *item)
{
	setupValueChanges(item);
	emit layoutChanged();
}

void VeQItemTableModel::onItemAboutToBeRemoved(VeQItem *item)
{
	mMap.remove(item->uniqueId());
	emit layoutChanged();
}

// "properties" / custom roles when the model is used for Qt Quick
QHash<int, QByteArray> VeQItemTableModel::roleNames() const
{
	QHash<int, QByteArray> roles;

	roles[IdRole] = "id";
	roles[UniqueIdRole] = "uid";
	roles[ValueRole] = "value";
	roles[StateRole] = "state";
	roles[TextRole] = "text";
	roles[TextStateRole] = "textState";

	return roles;
}

// return the actual data to be displayed for index
QVariant VeQItemTableModel::data(const QModelIndex &index, int role) const
{
	if (!index.isValid())
		return QVariant();

	VeQItem *item = static_cast<VeQItem*>(index.internalPointer());
	if (role == Qt::DisplayRole) {
		QString column = mColumns[index.column()];
		if (column == "id")
			role = UniqueIdRole;
		else if (column == "value")
			role = ValueRole;
		else if (column == "state")
			role = StateRole;
		else if (column == "text")
			role = TextRole;
		else if (column == "textState")
			role = TextStateRole;
		else
			return item->property(column.toLatin1());
	}

	switch (role) {
	case IdRole:
		return item->id();
	case UniqueIdRole:
		return item->uniqueId();
	case ValueRole:
		return item->getValue();
	case StateRole:
		return item->getState();
	case TextRole:
		return item->getText();
	case TextStateRole:
		return item->getTextState();
	}

	return QVariant();
}

Qt::ItemFlags VeQItemTableModel::flags(const QModelIndex &index) const
{
	if (!index.isValid())
		return 0;

	Qt::ItemFlags flags = Qt::ItemIsEnabled | Qt::ItemIsSelectable;

	if (index.column() == 1)
		flags |= Qt::ItemIsEditable;

	return flags;
}

bool VeQItemTableModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
	if (!index.isValid())
		return false;

	if (role == Qt::EditRole) {
		VeQItem *item = static_cast<VeQItem*>(index.internalPointer());
		item->setValue(value);
	}

	return true;
}

QVariant VeQItemTableModel::headerData(int section, Qt::Orientation orientation,
							   int role) const
{
	if (orientation == Qt::Horizontal && role == Qt::DisplayRole)
		return mColumns[section];

	return QVariant();
}

/*
 * Create a model index. Note the index can hold arbitrary pointer
 * available as internalPointer().
 */
QModelIndex VeQItemTableModel::index(int row, int column, const QModelIndex &parent) const
{
	Q_UNUSED(parent);

	if (row < 0 || row >= mMap.count())
		return QModelIndex();

	QString key(mMap.keys()[row]);
	VeQItem *item = mMap[key];

	return createIndex(row, column, item);
}

// Return the parent given an valid index.
QModelIndex VeQItemTableModel::parent(const QModelIndex &index) const
{
	Q_UNUSED(index);

	return QModelIndex();
}

// returns the number of rows (child count) of index
int VeQItemTableModel::rowCount(const QModelIndex &index) const
{
	if (index.column() > 0)
		return 0;

	int count = mMap.count();
	return count;
}
