#include <QDebug>
#include <QtDBus>
#include <QtXml>

#include <velib/velib_config.h>
#include <velib/qt/ve_qitems_dbus.hpp>
#include "v_busitem_proxy.h"

VeQItemDbus::VeQItemDbus(VeQItemDbusProducer *producer) :
	VeQItem(producer),
	mInterface(0),
	mRequestValueWhenOnline(false),
	mRequestTextWhenOnline(false),
	mRequestDefaultWhenOnline(false),
	mRequestMaxWhenOnline(false),
	mRequestMinWhenOnline(false)
{
}

VeQItemDbusProducer *VeQItemDbus::producer()
{
	return static_cast<VeQItemDbusProducer *>(mProducer);
}

/*
 * qtDbus creates a variant of QDBusArgument for all none basic types
 * encountered in case of a QDBusVariant. These can be be casted / demarshalled to
 * an explicit type. However qml will not perform such casts and complains it can't
 * deal with a QDBusArgument. In order to make this work, the casts are performed
 * here. This is in no way complete. QVariantMap and arrays of basic types are supported.
 *
 * NOTE: The remote c side has no onChange support at the moment (june 2014)
 * NOTE: I don't think these types are used anymore (august 2015)
 */
void VeQItemDbus::demarshallVariantForQml(QVariant &variant, QString &signature)
{
	QString dummy;

	if (variant.canConvert<QDBusArgument>())
	{
		const QDBusArgument &dbusArgument = qvariant_cast<QDBusArgument>(variant);
		signature = dbusArgument.currentSignature();
		switch (dbusArgument.currentType())
		{
		case QDBusArgument::MapType:
			/* QVariantMap support */
			if (dbusArgument.currentSignature() == "a{sv}")
			{
				QVariantMap map = qdbus_cast<QVariantMap>(dbusArgument);
				for (QVariantMap::iterator i = map.begin(); i != map.end(); i++)
					demarshallVariantForQml(i.value(), dummy);
				variant = QVariant::fromValue(map);
			} else {
				qDebug() << "unsupported MapType, your likely toasted...";
			}
			break;

		case QDBusArgument::ArrayType:
		{
			// An empty array represents an invalid value for busitems,
			// simply because there is no simple method to invalidate
			// a value on the dbus at the moment.
			dbusArgument.beginArray();
			if (dbusArgument.atEnd()) {
				dbusArgument.endArray();
				variant = QVariant();
				signature = "";
				break;
			}

			// Upcast everything to a QVariantList as qml understands that
			// by default..
			QVariantList list;
			while (!dbusArgument.atEnd())
				list.append(dbusArgument.asVariant());
			dbusArgument.endArray();

			for (QVariantList::iterator i = list.begin(); i != list.end(); i++)
				demarshallVariantForQml(*i, dummy);

			variant = QVariant::fromValue(list);
			break;
		}

		default:
			qDebug() << "unsupported QDBusArgument, your likely toasted...";
		}
	} else {
		signature = QDBusMetaType::typeToSignature(variant.type());
	}
}

void VeQItemDbus::onPropertiesChanged(const QVariantMap &changes)
{
	QMapIterator<QString, QVariant> i(changes);
	VeQItem *item = this;
	QString dummy;

	if (calledFromDBus() && !producer()->getListenIndividually())
		item = itemGetOrCreate(message().path());

	while (i.hasNext()) {
		i.next();
		if (i.key() == "Value") {
			QVariant value(i.value());
			demarshallVariantForQml(value, mSignature);
			item->produceValue(value);

		} else if (i.key() == "Text") {
			QString value(i.value().toString());
			item->produceText(value);

		} else if (i.key() == "Max") {
			QVariant value(i.value());
			demarshallVariantForQml(value, dummy);
			item->itemProduceProperty("max", value);

		} else if (i.key() == "Min") {
			QVariant value(i.value());
			demarshallVariantForQml(value, dummy);
			item->itemProduceProperty("min", value);

		} else if (i.key() == "Default") {
			QVariant value(i.value());
			demarshallVariantForQml(value, dummy);
			item->itemProduceProperty("defaultValue", value);
		}
	}
}

void VeQItemDbus::serviceRegistrationChanged(bool registered)
{
	if (registered) {
		/*
		 * Since the ccgx gui does not look at state changes at the moment, a state
		 * transition goes unnoticed. When a service invalidates an item, disconnects
		 * and returns without sending a change event itself, there won't be a pending
		 * action, since there was no value change in the first place (so getValue won't
		 * get called, since it didn't change, and hence didn't get marked as pending).
		 * For now just sync both text and value if there is a consumer of this item.
		 */
		if (mWatched) {
			mRequestValueWhenOnline = true;
			mRequestTextWhenOnline = true;
		}

		if (mRequestValueWhenOnline) {
			mRequestValueWhenOnline = false;
			getValue(true);
		}
		if (mRequestTextWhenOnline) {
			mRequestTextWhenOnline = false;
			getText(true);
		}
		if (mRequestDefaultWhenOnline) {
			mRequestDefaultWhenOnline = false;
			itemProperty("defaultValue", true);
		}
		if (mRequestMaxWhenOnline) {
			mRequestMaxWhenOnline = false;
			itemProperty("max", true);
		}
		if (mRequestMinWhenOnline) {
			mRequestMinWhenOnline = false;
			itemProperty("min", true);
		}

		// If there was no pending action at all mark as Idle
		if (mState == Offline)
			setState(Idle);
		if (mTextState == Offline)
			setTextState(Idle);

	} else {
#if CFG_DBUS_INVALIDATE_OFFLINE
		/*
		 * Ideally the values are kept but only the state is changed, so the user interface
		 * has a choice how to cope with it. Since the ccgx gui doesn't do that at the moment,
		 * invalidate it so the display gets updated accordingly.
		 */
		produceValue(QVariant(), Offline);
		produceText("", Offline);
#else
		setState(Offline);
		setTextState(Offline);
#endif
	}
}

void VeQItemDbus::introspectObtained(QDBusPendingCallWatcher *call)
{
	QDBusPendingReply<QString> reply = *call;
	QDomDocument doc;

	if (!reply.isValid()) {
		qDebug() << "Cannot introspect object" <<  reply.error();
	} else {
		doc.setContent(reply.value());
		QDomElement child = doc.firstChildElement().firstChildElement();
		while (!child.isNull()) {
			if (child.tagName() == "node") {
				QString name = child.attribute("name");
				itemGetOrCreate(name);
			}

			child = child.nextSiblingElement();
		}
	}

	call->deleteLater();
}

class FreeDesktopIntrospect : public QDBusAbstractInterface
{
public:
	FreeDesktopIntrospect(const QString &service, const QString &path,
						   const QDBusConnection &connection, QObject *parent = 0) :
		QDBusAbstractInterface(service, path, "org.freedesktop.DBus.Introspectable", connection, parent)
	{
	}
};

bool VeQItemDbus::introspect()
{
	FreeDesktopIntrospect iface(mInterface->service(), mInterface->path(), mInterface->connection());

	if (!iface.isValid()) {
		qDebug() << "Cannot introspect object" <<  mInterface->path();
		return false;
	}

	QList<QVariant> argumentList;
	QDBusPendingReply<QString> call = iface.asyncCallWithArgumentList("Introspect", argumentList);
	QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(call, this);
	this->connect(watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), SLOT(introspectObtained(QDBusPendingCallWatcher*)));

	return true;
}

void VeQItemDbus::valueObtained(QDBusPendingCallWatcher *call)
{
	QDBusPendingReply<QDBusVariant> reply = *call;
	QVariant value;
	State state(Synchronized);

	if (reply.isError()) {
		QDBusError error(reply.error());
		state = Offline;
		// tolerate paths not yet being present and failing GetValues on the root
		if (error.type() != QDBusError::UnknownObject && mInterface->path() != "/")
			qDebug() << "Error value:" << reply.error() << mInterface->service();
	} else {
		value = reply.value().variant();
		demarshallVariantForQml(value, mSignature);

		if (!isLeaf()) {
			// when support, which is apperently is, having a valid responds on a non-leaf
			// item returns all the values of the leaves in one go. So iterate and update them
			// all as well.
			QVariantMap map(value.toMap());
			for (QVariantMap::const_iterator i = map.constBegin(); i != map.constEnd(); i++)
				itemGetOrCreateAndProduce(i.key(), i.value());
			// Note: only the leaves receive change events, so clear the actual value
			// to prevent users from using the non-leaves
			value = QVariant();
		}
	}

	produceValue(value, state);
	call->deleteLater();
}

QVariant VeQItemDbus::getValue(bool force)
{
	if (mState == Idle || force) {
		// If the service is not online, postpone the request till it is ...
		if (!producer()->isServiceRegistered(mInterface->service())) {
			mRequestValueWhenOnline = true;
			setState(Offline);
			return QVariant();
		}

		QList<QVariant> argumentList;
		setState(Requested);
		QDBusPendingReply<QDBusVariant> call = mInterface->asyncCallWithArgumentList("GetValue", argumentList);
		QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(call, this);
		this->connect(watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), SLOT(valueObtained(QDBusPendingCallWatcher*)));
	}

	return mValue;
}

void VeQItemDbus::textObtained(QDBusPendingCallWatcher *call)
{
	QDBusPendingReply<QString> reply = *call;
	QString value;
	State state(Synchronized);

	if (reply.isError()) {
		QDBusError error(reply.error());
		state = Offline;
		// tolerate paths not yet being present
		if (error.type() != QDBusError::UnknownObject)
			qDebug() << "Error value:" << reply.error() << mInterface->service();
	} else {
		value = reply.value();
	}

	produceText(value, state);
	call->deleteLater();
}

QString VeQItemDbus::getText(bool force)
{
	if (mTextState == Idle || force) {
		// If the service is not online, postpone the request till it is ...
		if (!producer()->isServiceRegistered(mInterface->service())) {
			mRequestTextWhenOnline = 1;
			setTextState(Offline);
			return QString();
		}

		QList<QVariant> argumentList;
		setTextState(Requested);
		QDBusPendingReply<QString> call = mInterface->asyncCallWithArgumentList("GetText", argumentList);
		QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(call, this);
		this->connect(watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), SLOT(textObtained(QDBusPendingCallWatcher*)));
	}

	return mText;
}

void VeQItemDbus::propertyObtained(const char *name, QDBusPendingCallWatcher *call)
{
	QDBusPendingReply<QDBusVariant> reply = *call;
	QVariant value;
	State state(Synchronized);

	if (reply.isError()) {
		QDBusError error(reply.error());
		state = Offline;
		// tolerate paths not yet being present
		if (error.type() != QDBusError::UnknownObject)
			qDebug() << "Error value:" << reply.error() << mInterface->service();
	} else {
		value = reply.value().variant();
		QString dummy;
		demarshallVariantForQml(value, dummy);
	}

	itemProduceProperty(name, value, state);
	call->deleteLater();
}

void VeQItemDbus::defaultObtained(QDBusPendingCallWatcher *call)
{
	propertyObtained("defaultValue", call);
}

void VeQItemDbus::maxObtained(QDBusPendingCallWatcher *call)
{
	propertyObtained("max", call);
}

void VeQItemDbus::minObtained(QDBusPendingCallWatcher *call)
{
	propertyObtained("min", call);
}

QVariant VeQItemDbus::itemProperty(const char *name, bool force)
{
	if (mPropertyState[name] == Idle || force) {
		QString method;
		bool *pending;
		const char *slot;

		if (QLatin1String(name) == "min") {
			method = "GetMin";
			pending = &mRequestMinWhenOnline;
			slot = SLOT(minObtained(QDBusPendingCallWatcher*));
		} else if (QLatin1String(name) == "max") {
			method = "GetMax";
			pending = &mRequestMaxWhenOnline;
			slot = SLOT(maxObtained(QDBusPendingCallWatcher*));
		} else if (QLatin1String(name) == "defaultValue") {
			method = "GetDefault";
			pending = &mRequestDefaultWhenOnline;
			slot = SLOT(defaultObtained(QDBusPendingCallWatcher*));
		} else {
			return property(name);
		}

		// If the service is not online, postpone the request till it is ...
		if (!producer()->isServiceRegistered(mInterface->service())) {
			*pending = true;
			mPropertyState[name] = Offline;
			return QVariant();
		}

		QList<QVariant> argumentList;
		mPropertyState[name] = Requested;
		QDBusPendingReply<QDBusVariant> call = mInterface->asyncCallWithArgumentList(method, argumentList);
		QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(call, this);
		this->connect(watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), slot);
	}

	return property(name);
}

void VeQItemDbus::setValueDone(QDBusPendingCallWatcher *call)
{
	call->deleteLater();
}

int VeQItemDbus::setValue(const QVariant &value)
{
	if (!producer()->isServiceRegistered(mInterface->service()) || mState == Offline) {
		qDebug() << "ignoring request for" << uniqueId() << "to set value" << value << "(not online)";
		return -1;
	}

	// no need to set values to the same again
	if (mState == Storing && mPendingSetValue == value)
		return 0;

	mPendingSetValue = value;
	setState(Storing);

	QDBusPendingReply<int> set = mInterface->SetValue(QDBusVariant(value));
	QDBusPendingCallWatcher *watcher = new QDBusPendingCallWatcher(set, this);
	this->connect(watcher, SIGNAL(finished(QDBusPendingCallWatcher*)), SLOT(setValueDone(QDBusPendingCallWatcher*)));

	return 0;
}

// hook just before the item is added
void VeQItemDbus::setParent(QObject *parent)
{
	VeQItem::setParent(parent);
	Q_ASSERT(mInterface == 0);
	QString service;
	QString path;

	// If this is a dbus service check if it is known
	VeQItemDbus *parentQItem = static_cast<VeQItemDbus *>(parent);
	if (parentQItem == producer()->services()) {
		service = mId;
		path = "/";
		producer()->serviceWatchAdd(service);
	} else {
		service = parentQItem->mInterface->service();
		path = parentQItem->mInterface->path() + (parentQItem->mInterface->path() == "/" ? "" : "/") + mId;
	}

	mInterface = new VBusItemProxy(service, path, producer()->dbusConnection(), this);

	// special treatment of the root, it handles all change events.
	// It might report all initial values at once for init when enabled
	if (producer()->getListenIndividually()) {
		Q_ASSERT(!producer()->getBulkInit());
		if (mIsLeaf && path != "/") {
			mInterface->connection().connect(service, path, "com.victronenergy.BusItem" ,"PropertiesChanged",
											 this, SLOT(onPropertiesChanged(const QVariantMap &)));
		}
	} else if (path == "/") {
		mInterface->connection().connect(service, "", "com.victronenergy.BusItem" ,"PropertiesChanged",
										 this, SLOT(onPropertiesChanged(const QVariantMap &)));
		if (producer()->getBulkInit())
			getValue();
	}
}

/**
 * @brief VeQItemDbusProducer::VeQItemDbusProducer
 * @param root
 * @param id
 * @param findVictronServices
 *          There are two modus operandi. Well known services can be added and monitored for
 *          presence. com.victronenergy services typically have "something" unique added
 *          to the dbus service name to allow multiple service of the same kind (e.g. multiple
 *          battery banks, motors etc). Setting this flag will auto add all services starting
 *          with com.victronengy. (It will cause a warning to be displayed about connecting to
 *          a depecrated signal to monitor all services, which is exactly the intended, the warning
 *          is the BUG)
 * @param bulkInitOfNewService
 *          Some services (c based at the moment) allow the get all values at once
 *          by calling GetValue on its root item. This can be enabled to initialize,
 *          but will complain when querying services which don't support it. When
 *          debugging you typically want to turn this off.
 * @param parent
 */
VeQItemDbusProducer::VeQItemDbusProducer(VeQItem *root, QString id,
										 bool findVictronServices, bool bulkInitOfNewService,
										 QObject *parent)
	: VeQItemProducer(root, id, parent),
	  mDbus(QDBusConnection("")),
	  mFindVictronServices(findVictronServices),
	  mBulkInitOfNewService(bulkInitOfNewService),
	  mListenIndividually(false)
{
}

VeQItemDbusProducer::~VeQItemDbusProducer()
{
	// guess it is https://bugreports.qt.io/browse/QTBUG-29498
	if (!getBulkInit() && getListenIndividually()) {
		foreach (VeQItem *item, mProducerRoot->itemChildren())
			disconnectChildren(item, item);
	} else {
		foreach (VeQItem *item, mProducerRoot->itemChildren())
			dbusConnection().disconnect(item->id(), "", "com.victronenergy.BusItem" ,"PropertiesChanged", item, SLOT(onPropertiesChanged(const QVariantMap &)));
	}
}

bool VeQItemDbusProducer::open(const QString &address, const QString &qtDbusName)
{
	// FIXME: find non blocking version, all signals / slots are stalled for
	// 30 secs or so if this fails when connection by tcp/ip...
	if (address == "session")
		return open(QDBusConnection::connectToBus(QDBusConnection::SessionBus, qtDbusName));
	if (address == "system")
		return open(QDBusConnection::connectToBus(QDBusConnection::SystemBus, qtDbusName));

	return open(QDBusConnection::connectToBus(address, qtDbusName));
}

bool VeQItemDbusProducer::open(const QDBusConnection &dbusConnection)
{
	mDbus = dbusConnection;

	if (!mDbus.isConnected())
		return false;

	if (mFindVictronServices) {
		QDBusConnectionInterface *interface = mDbus.interface();
		connect(interface, SIGNAL(serviceOwnerChanged(QString,QString,QString)), this, SLOT(onServiceOwnerChanged(QString,QString,QString)));

		QStringList registerServices = interface->registeredServiceNames();
		foreach (QString name, registerServices) {
			if (name.startsWith("com.victronenergy."))
				mProducerRoot->itemGetOrCreate(name, false);
		}
	}

	return true;
}

VeQItem *VeQItemDbusProducer::createItem()
{
	return new VeQItemDbus(this);
}

void VeQItemDbusProducer::serviceWatchAdd(QString serviceName)
{
	Q_ASSERT(mDbus.interface());

	QDBusServiceWatcher *watcher = new QDBusServiceWatcher(serviceName, mDbus);
	connect(watcher, SIGNAL(serviceRegistered(QString)), SLOT(onServiceRegistered(QString)));
	connect(watcher, SIGNAL(serviceUnregistered(QString)), SLOT(onServiceUnregistered(QString)));

	QStringList registeredServices(mDbus.interface()->registeredServiceNames());
	bool registered = registeredServices.contains(serviceName);
	mServiceRegistered.insert(serviceName, registered);
	mServiceWatchers.insert(serviceName, watcher);
}

/**
 * Listen to each requested D-Bus object individually for changes.
 * Setting this option to true will improve performance if you are interested in a few D-Bus items
 * per service.
 */
void VeQItemDbusProducer::setListenIndividually(bool l)
{
	if (l && mBulkInitOfNewService) {
		qDebug() << "ignoring 'listen individually' flag, because bulk init is enabled";
		return;
	}
	if (mDbus.isConnected()) {
		qDebug() << "cannot change 'listen individually' flag when the producer has already been opened";
		return;
	}
	mListenIndividually = l;
}

void VeQItemDbusProducer::onServiceRegistered(const QString &service)
{
	onServiceRegistrationChanged(service, true);
}

void VeQItemDbusProducer::onServiceUnregistered(const QString &service)
{
	onServiceRegistrationChanged(service, false);
}

bool VeQItemDbusProducer::isServiceRegistered(QString serviceName)
{
	return mServiceRegistered.value(serviceName);
}

void VeQItemDbusProducer::onServiceRegistrationChanged(const QString &serviceName, bool registered)
{
	mServiceRegistered.insert(serviceName, registered);
	VeQItemDbus *item = static_cast<VeQItemDbus *>(mProducerRoot->itemGet(serviceName));
	Q_ASSERT(item);
	item->serviceRegistrationChanged(registered);
	foreach(VeQItemDbus *child, item->findChildren<VeQItemDbus *>())
		child->serviceRegistrationChanged(registered);
}

void VeQItemDbusProducer::disconnectChildren(VeQItem *root, VeQItem *item)
{
	QString serviceName = root->id();
	foreach (VeQItem *child, item->itemChildren()) {
		if (child->isLeaf()) {
			dbusConnection().disconnect(serviceName, child->getRelId(root),
										"com.victronenergy.BusItem" ,"PropertiesChanged",
										child, SLOT(onPropertiesChanged(const QVariantMap &)));
		} else {
			disconnectChildren(root, child);
		}
	}
}

// Finds not "well known services"
void VeQItemDbusProducer::onServiceOwnerChanged(const QString &serviceName, const QString &oldOwner, const QString &newOwner)
{
	Q_UNUSED(oldOwner);

	if (!serviceName.startsWith("com.victronenergy."))
		return;

	if (newOwner != "")
		mProducerRoot->itemGetOrCreate(serviceName, false);
}
